#include <stdio.h>
#include <string.h>
#include <ctype.h>

#define FALSE	0
#define TRUE    !FALSE

extern void Error(char *Format, ...);
extern int MatchDotFlag;
extern int MatchSlashFlag;

/**************************************************************************
 * AMatch -- Return TRUE if 'Str' matches 'Pat'.  Based on the recursive
 * routine AMatch() (for "anchored match") in Kernighan & Plauger's
 * "Software Tools".
 *************************************************************************/
int
AMatch(char *Str, char *Pat) {
	int Sc, Pc;			/* character at current position */

	while ((Pc = *Pat++) != '\0') {
		if (Pc == '[') {
			Sc = *Str++;
			for (;;) {
				if (*Pat == ']') return FALSE;
				if (Sc == *Pat++) break;
				}
			while (*Pat != ']' && *Pat != '\0') Pat++;
			if (*Pat != ']') Error("Pattern missing closing \"]\"");
			}
		else if (Pc == '*') {
			int NChars;

			/* Try to match as many characters as possible. */

			NChars = 0;
			while ((Sc = *Str) != '\0') {
				if (Sc == '.' && !MatchDotFlag) break;
				if (strchr("/\\", Sc) != NULL && !MatchSlashFlag) break;
				Str++;
				NChars++;
				}

			/* We are now at the character that made us
			 * fail.  Try to match the rest of the pattern.
			 * Shrink the closure by one for each failure.
			 * Since closure matches *zero* or more occurences
			 * of a pattern, a match may start even if the
			 * previous loop matched no characters.
			 */
			for (;;) {
				if (AMatch(Str--, Pat)) return (TRUE);
				if (NChars-- == 0) return (FALSE);
				}
			}
			/* ? matches anything, but we only advance if not null. */
		else if (Pc == '?' && *Str == '.') {
			if (MatchDotFlag) Str++;
			}
		else if (Pc == '?' && strchr("/\\", *Str) != NULL) {
			if (MatchSlashFlag) Str++;
			}
		else if (Pc == '?' && *Str == '\0') ;
		else if (Pc == '?') Str++;
			/* exit if they don't match. */
		else if (toupper(Pc) != toupper(*Str++)) return (FALSE);
		}			/* End of *Str++ loop.*/
	return (*Str == '\0');
	}