#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <conio.h>
#include <dos.h>
#include <io.h>
#include <fcntl.h>
#include <sys\types.h>
#include <sys\stat.h>
#ifdef __TURBOC__
	#include <alloc.h>
	#include <dir.h>
#else
	#ifdef __POWERC__
		#include <alloc.h>
		#include <dir.h>
	#else
		#ifdef M_I86
			#include <malloc.h>
			#include <direct.h>
		#endif
	#endif
#endif

#define TRUE 1
#define FALSE 0

#include "queue.h"
#include "ufind.h"
#include "copymove.h"

/*--- Function Prototypes  ---*/

/* -From crc.c                */

unsigned int ComputeCRC(unsigned int CRCSum, int Count, char *Buf);

int GetYes (char *Format, ...);


/* return a string.  If the string begins with '?', it's an error message. */
char *
CopyMove (char *FromFn, char *ToFn, int MoveFlag, int VerifyFlag) {

	int FromFD, ToFD;
	int Found;
	struct DIR_BLK DirBlk;
	char *BufferPtr;
	unsigned BufferSize;
	int ReadCount, WriteCount;
	unsigned OldCRC, NewCRC;

	union {
		unsigned long I;
#ifdef __TURBOC__
		struct ftime S;
#else
	#ifdef __POWERC__
		struct ftime S;
	#else
		#ifdef M_I86
		struct {
			unsigned f_time;
			unsigned f_date;
			} V;
		#endif
	#endif
#endif
		} SrcDate, DestDate;

	char *ErrMsg;

	FromFD = open(FromFn, O_RDONLY | O_BINARY);
	if (FromFD < 0) {
		ErrMsg = "?Can't open input file";
		goto AbortNotOpen;
		}

/* see if the destination exists */

	if ( (Found = !FIND_FIRST(ToFn, &DirBlk, DIREC_)) != 0) {

/* see if we can/should overwrite an existing file */

		if (DirBlk.DIR_ATTR & DIREC_) {
			ErrMsg = "?Can't overwrite directory";
			goto AbortOpen;
			}
		else if ( NewFlag ) {
			ToFD = open(ToFn, O_RDONLY | O_BINARY);
#ifdef __TURBOC__
			getftime(FromFD, &SrcDate.S);
			getftime(ToFD, &DestDate.S);
#else
	#ifdef __POWERC__
			getftime(FromFD, &SrcDate.S);
			getftime(ToFD, &DestDate.S);
	#else
		#ifdef M_I86
			_dos_getftime(FromFD, &SrcDate.V.f_date, &SrcDate.V.f_time);
			_dos_getftime(ToFD, &DestDate.V.f_date, &SrcDate.V.f_time);
		#endif
	#endif
#endif
			close(ToFD);
			if (DestDate.I >= SrcDate.I) goto AbortOpen;
			}
		else if (WarnFlag && !GetYes("Destination file:%s exists. Overwrite", ToFn))
			goto AbortOpen;
		}

/* delete the destination */

	if ( !YesNo("%s %s TO %s", MvFlag ? "MV" : "CP", FromFn, ToFn) )
		goto AbortOpen;
	if (Found && unlink(ToFn) != 0) {
		ErrMsg = "?Unable to overwrite file";
		goto AbortOpen;
		}

/* if we're just moving it someplace else on the same drive, do so. */
/* assume that the filenames are fully specified, including drive */

	if (MoveFlag && *FromFn == *ToFn) {
		close(FromFD);	/* we can only rename a closed file */
		if (rename(FromFn, ToFn)) ErrMsg = "?Unable to move file";
		else ErrMsg = "Moved";
		goto AbortNotOpen;
		}

/* if we're moving it to a different drive, turn verify on */

	if (MoveFlag) VerifyFlag = TRUE;

/* allocate memory for the buffer */

#ifdef __TURBOC__
	BufferSize = coreleft() - 100;
#else
	#ifdef __POWERC__
	BufferSize = coreleft() - 100;
	#else
		#ifdef M_I86
	BufferSize = _memavl() - 100;
		#endif
	#endif
#endif
	if (BufferSize < 512) {
		ErrMsg = "?Insufficient memory";
		goto AbortNotOpen;
		}
	BufferSize = min(BufferSize, 0xFA00);
	if ( (BufferPtr = malloc(BufferSize)) == NULL) goto AbortOpen;;

/* create the destination */

	ToFD = open(ToFn, O_WRONLY | O_CREAT | O_BINARY, S_IREAD | S_IWRITE);
	if (ToFD < 0) {
		ErrMsg = "?Can't create output file";
		goto AbortMemory;
		}

	OldCRC = 0;

	while ((ReadCount = read(FromFD, BufferPtr, BufferSize)) != 0) {
		if (ReadCount == -1) {
			close(ToFD);
			unlink(ToFn);
			ErrMsg = "?Input read error";
			goto AbortMemory;
			}
		if (VerifyFlag) OldCRC = ComputeCRC(OldCRC, ReadCount, BufferPtr);
		WriteCount = write(ToFD, BufferPtr, ReadCount);
		if (WriteCount == -1 || WriteCount != ReadCount) {
			close(ToFD);
			unlink(ToFn);
			ErrMsg = "?Disk Full";
			goto AbortMemory;
			}
		}

/* set outf's timestamp equal to inf's timestamp */

	{
#ifdef __TURBOC__
		struct ftime When;

		getftime(FromFD, &When);
		setftime(ToFD, &When);
#else
	#ifdef __POWERC__
		struct ftime When;

		getftime(FromFD, &When);
		setftime(ToFD, &When);
	#else
		#ifdef M_I86
		unsigned f_time;
		unsigned f_date;

		_dos_getftime(FromFD, &f_date, &f_time);
		_dos_setftime(ToFD, f_date, f_time);
		#endif
	#endif
#endif
	}

	close(ToFD);

/* verify that the output file was correct if desired */

	if (VerifyFlag) {	/* reset the disk system lest we get it from buffers. */
		union REGS Regs;

		Regs.h.ah = 0x0D;
		int86(0x21, &Regs, &Regs);

		ToFD = open(ToFn, O_RDONLY | O_BINARY);
		if (ToFD < 0) {
			ErrMsg = "?Can't open output file for reading";
			goto AbortMemory;
			}

		NewCRC = 0;

		while ((ReadCount = read(ToFD, BufferPtr, BufferSize)) != 0) {
			if (ReadCount == -1) {
				close(ToFD);
				ErrMsg = "?Output read error";
				goto AbortMemory;
				}
			NewCRC = ComputeCRC(NewCRC, ReadCount, BufferPtr);
			}
		close(ToFD);
		if (OldCRC != NewCRC) {
			close(ToFD);
			ErrMsg = "?Error on CRC compare";
			goto AbortMemory;
			}
		}

/* if we're moving it to a different drive, delete the original. */

	if (MoveFlag && unlink(FromFn)) {
		ErrMsg = "?Unable to delete original copy";
		goto AbortMemory;
		}

/* report success */

	if (VerifyFlag) ErrMsg = "Copied";
	else ErrMsg = "CRC Verified";

AbortMemory:
	free(BufferPtr);
AbortOpen:
	close(FromFD);
AbortNotOpen:
	return (ErrMsg);
	}

int
GetYes (char *Format, ...) {
	va_list ArgPtr;
	int Reply;

	while (1) {
		va_start(ArgPtr, Format);
		vfprintf(stderr, Format, ArgPtr);
		fprintf(stderr, " (Y/N)? ");
		Reply = toupper(getch());
		if ( (Reply == 'Y') || (Reply == 'N') ) {
			va_end(ArgPtr);
			break;
			}
		fprintf(stderr, " %c is invalid \x07\n", Reply);
		continue;
		}
	printf(" %c\n", Reply);
	return( Reply == 'Y' );
	}


