#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <dos.h>
#ifdef __TURBOC__
	#include <dir.h>
#else
	#ifdef __POWERC__
		#include <dir.h>
	#else
		#ifdef M_I86
			#include <direct.h>
		#endif
	#endif
#endif

#undef MAIN

#include "queue.h"
#include "ufind.h"

void GetDestDir (char *InPtr, char **argv, char *ToDir);
int GetCopyMove (char *Ptr);
void Usage (void);

static char *OpTable[] = {
#define S_NAME 0
	"name",
#define S_NAME_N         1
	"^name",
#define S_FNAME          2
	"fname",
#define S_FNAME_N        3
	"^fname",
#define S_ATTR           4
	"attr",
#define S_ATTR_N         5
	"^attr",
#define S_LS             6
	"ls",
#define S_LD             7
	"ld",
#define S_PRINT          8
	"print",
#define S_ATTR_ON        9
	"attr+",
#define S_ATTR_OFF      10
	"attr-",
#define S_NR            11
	"nr",
#define S_SIZE          12
	"size",
#define S_REMOVE        13
	"rm",
#define S_BACKUP        14
	"backup",
#define S_REMOVE_DIR    15
	"rmdir",
#define S_SUM           16
	"sum",
#define S_COMP          17
	"comp",
#define S_COMP_N        18
	"^comp",
#define S_EXEC          19
	"exec",
#define S_LT_DATE       20
	"ltdate",
#define S_LE_DATE       21
	"ledate",
#define S_GT_DATE       22
	"gtdate",
#define S_GE_DATE       23
	"gedate",
#define S_LESS_THAN     24
	"lt",
#define S_LESS_EQUAL    25
	"le",
#define S_GREATER_THAN  26
	"gt",
#define S_GREATER_EQUAL 27
	"ge",
#define S_SINCE         28
	"si",
#define S_BEFORE        29
	"be",
#define S_EXISTS        30
	"exists",
#define S_EXISTS_N      31
	"exists^",
#define S_NCOMP			32
	"compn",
#define S_NCOMP_N		33
	"^compn",
#define S_UNKNOWN  		34
	""
	};

/**************************************************************************
 * GetArgs - Get and interpret command line arguments
 *************************************************************************/
void
GetArgs (char ***argv) {
	char *InPtr, Temp[66];
	static char OnlyOneErr[] = "must pick only one of mv, cp and backup";
	int i;
	struct DIR_BLK ThisFF;

	(*argv)++;
	while ((InPtr = *(*argv)) != NULL && *InPtr != SwitChar) {
		RootPath(InPtr, Temp);
		Enque(&PNames, Temp);
		(*argv)++;
		}

	while ((InPtr = *(*argv)++) != NULL) {
		if (*InPtr != SwitChar)
			Error("You specified a pathname (%s) after an option", InPtr);

/* File selection criteria */

		if (InPtr[1] == 'c'&& strlen(InPtr) < 4) {
			if (MvFlag || BackupFlag) Error(OnlyOneErr);
			CpFlag = TRUE;
			if ( GetCopyMove(InPtr) ) {
				GetDestDir(*(*argv)++, *argv, ToDir);
				continue;
				}
			}
		else if (InPtr[1] == 'm') {
			if (CpFlag || BackupFlag) Error(OnlyOneErr);
			MvFlag = TRUE;
			if ( GetCopyMove(InPtr) ) {
				GetDestDir(*(*argv)++, *argv, ToDir);
				continue;
				}
			}
		for (i = 0; OpTable[i][0] != '\0'; i++) {
			if (stricmp(&InPtr[1], OpTable[i]) == 0) break;
			}
		switch (i) {
			case S_NAME:
				while ((InPtr = *(*argv)) != NULL && *InPtr != SwitChar) {
					if (IncNamesCnt >= sizeof(IncNames) / sizeof(char *))
						Error("name: too many filenames");
					IncNames[IncNamesCnt++] = InPtr;
					(*argv)++;
					}
				break;
			case S_NAME_N:
				while ((InPtr = *(*argv)) != NULL && *InPtr != SwitChar) {
					if (ExcNamesCnt >= sizeof(ExcNames) / sizeof(char *))
						Error("name: too many filenames");
					ExcNames[ExcNamesCnt++] = InPtr;
					(*argv)++;
					}
				break;
			case S_FNAME:
				while ((InPtr = *(*argv)) != NULL && *InPtr != SwitChar) {
					if (CntIncFNames >= sizeof(IncFNames) / sizeof(char *))
						Error("fname: too many filenames");
					IncFNames[CntIncFNames++] = InPtr;
					(*argv)++;
					}
				break;
			case S_FNAME_N:
				while ((InPtr = *(*argv)) != NULL && *InPtr != SwitChar) {
					if (CntExcFNames >= sizeof(ExcFNames) / sizeof(char *))
						Error("fname: too many filenames");
					ExcFNames[CntExcFNames++] = InPtr;
					(*argv)++;
					}
				break;
			case S_ATTR:
				if ((InPtr = *(*argv)++) == NULL || *InPtr == SwitChar)
					Error("attr: must be followed by a list of attributes");
				IsAttrOn |= ParseAttributes(InPtr);
				break;
			case S_ATTR_N:
				if ((InPtr = *(*argv)++) == NULL || *InPtr == SwitChar)
					Error("^attr: must be followed by a list of attributes");
				IsAttrOff |= ParseAttributes(InPtr);
				break;
			case S_LESS_THAN:
			case S_LESS_EQUAL:
			case S_GREATER_THAN:
			case S_GREATER_EQUAL:
			case S_LT_DATE:
			case S_LE_DATE:
			case S_GT_DATE:
			case S_GE_DATE:
				if ((InPtr = *(*argv)++) == NULL || *InPtr == SwitChar)
					Error("time comparison: must be followed by a file or path name");
				if (FIND_FIRST(DtName = InPtr, &ThisFF, DIREC_))
					Error("time comparison: Cannot find '%s'", InPtr);
				DtTime = ThisFF.DIR_TIME;
				DtDate = ThisFF.DIR_DATE;
				DtDir = IsDir(ThisFF.DIR_ATTR) && FIND_NEXT(&ThisFF);
				switch(i) {
					case S_LESS_THAN: DtStyle = LT; break;
					case S_LESS_EQUAL: DtStyle = LE; break;
					case S_GREATER_THAN: DtStyle = GT; break;
					case S_GREATER_EQUAL: DtStyle = GE; break;
					case S_LT_DATE: DtStyle = LTDATE; break;
					case S_LE_DATE: DtStyle = LEDATE; break;
					case S_GT_DATE: DtStyle = GTDATE; break;
					case S_GE_DATE: DtStyle = GEDATE; break;
					}
				break;
			case S_SINCE:
				if ((*argv)[0] == NULL || (*argv)[0][0] == SwitChar) SiDate = Today;
				else {
					InPtr = *(*argv)++;
					SiDate = GetDateArg(InPtr);
					}
				if (DtStyle == BE) DtStyle = SIBE;
				else DtStyle = SI;
				break;
			case S_BEFORE:
				if ((*argv)[0] == NULL || (*argv)[0][0] == SwitChar) BeDate = Today;
				else {
					InPtr = *(*argv)++;
					BeDate = GetDateArg(InPtr);
					}
				if (DtStyle == SI) DtStyle = SIBE;
				else DtStyle = BE;
				break;
			case S_EXISTS:
			case S_EXISTS_N:
				if ((InPtr = *(*argv)++) == NULL || *InPtr == SwitChar)
					Error("exists: must be followed by a path name");
				if (FIND_FIRST(ExistsName = InPtr, &ThisFF, DIREC_)
					|| !IsDir(ThisFF.DIR_ATTR)
					|| !FIND_NEXT(&ThisFF)
					) Error("exists: Cannot find '%s'", InPtr);
				ExistsFlag = (i == S_EXISTS) ? 1 : -1;
				break;
			case S_SIZE:
				if ((InPtr = *(*argv)++) == NULL)
					Error("size: must be followed by an integer");
				if (*InPtr == '+' || *InPtr == SwitChar) SizeFlag = *InPtr++;
				else SizeFlag = '=';
				SizeSize = atol(InPtr);
				break;

/* actions */

			case S_LS:
				LsFlag = TRUE;
				break;
			case S_LD:
				LdFlag = TRUE;
				break;
			case S_PRINT:
				PrintFlag = TRUE;
				break;
			case S_ATTR_ON:
				if ((InPtr = *(*argv)++) == NULL || *InPtr == SwitChar)
					Error("attr+ must be followed by a list of attributes");
				AttrOn |= ParseAttributes(InPtr);
				break;
			case S_ATTR_OFF:
				if ((InPtr = *(*argv)++) == NULL || *InPtr == SwitChar)
					Error("attr- must be followed by a list of attributes");
				AttrOff |= ParseAttributes(InPtr);
				break;
			case S_BACKUP:
				if (CpFlag || MvFlag) Error(OnlyOneErr);
				if ((InPtr = *(*argv)++) == NULL || *InPtr == SwitChar)
					Error("BACKUP must be followed by a destination path");
				if (strchr("/\\", InPtr[strlen(InPtr)-1]) != NULL)
					sprintf(ToDir, "%s\\", InPtr);
				else strcpy(ToDir, InPtr);
				IsAttrOn |= ARCH;
				if ((AttrOn & ARCH) == 0) AttrOff |= ARCH;
				BackupFlag = TRUE;
				break;
			case S_REMOVE:
				RmFlag = TRUE;
				break;
			case S_REMOVE_DIR:
				RmDirFlag = TRUE;
				break;
			case S_SUM:
				SumFlag = TRUE;
				SumSize = 0L;
				break;
			case S_COMP:
			case S_COMP_N:
			case S_NCOMP:
			case S_NCOMP_N:
				if ((InPtr = *(*argv)++) == NULL || *InPtr == SwitChar)
					Error("comp must be followed by");
				if (strchr("/\\", InPtr[strlen(InPtr)-1]) != NULL)
					sprintf(CompDir, "%s\\", InPtr);
				else strcpy(CompDir, InPtr);
				CompFlag = (i == S_COMP || i == S_NCOMP) ? 1 : -1;
				NewSwt  = (i == S_NCOMP || i == S_NCOMP_N) ? 1 : 0;
				break;
			case S_EXEC:
				ExecFlag = TRUE;
				if ( ((InPtr = *(*argv)) == NULL) || (*InPtr == SwitChar) ) {
					Error("    -exec requires a command\n");
					}
				strcpy(CommandString, InPtr);
				(*argv)++;
				break;

/* options */

			case S_NR:
				NrFlag = TRUE;
				break;
			case S_UNKNOWN: {
				char *optionptr = InPtr;

				for (InPtr += 1; *InPtr != '\0'; InPtr++) {
					switch(*InPtr) {
						case 'a':
							IsAttrOn |= ARCH;
							break;
						case 's':
							QuietFlag = TRUE;
							break;
						case 'p':
							PromptFlag = TRUE;
							break;
						case 'n':
							NFlag = TRUE;
							break;
						case '.':
							MatchDotFlag = TRUE;
							break;
						case '/':
							MatchSlashFlag = TRUE;
							break;
						case 't':
							TFlag = TRUE;
							break;
						case 'w':
							WarnFlag = TRUE;
							break;
						default:
							fprintf(stderr, "unrecognized option '%c' in switch %s\nUSAGE: find pathname -switches\n",
								  *InPtr, optionptr);
							Usage();
							break;
						}
					}
				}
			}
		}
	}

void
GetDestDir (char *InPtr, char **argv, char *ToDir) {

	if ((InPtr = *argv) == NULL || *InPtr == SwitChar)
		Error("%s must be followed by a destination path",
			(MvFlag) ? "MV" : "CP");
	RootPath(InPtr, ToDir);
	}

int
GetCopyMove (char *Ptr) {
	char Ch;

	Ptr += 2;
	while ( (Ch = *(Ptr++)) != '\0') {
		switch (Ch) {
			case 'p':
				if (!CpFlag) return(0);
				break;
			case 'v':
				if (!MvFlag) return(0);
				break;
			case 't':
				if (MvFlag) RmDirFlag = TRUE;
				TFlag = TRUE;
				break;
			case 'n':
				NewFlag = TRUE;
				break;
			default:
				return(0);
			}
		}
	return(1);
	}

/**************************************************************************
 * GetDateArg - Get a date argument from command line and translate it
	 			into the format used by MS-DOS for files.
 *************************************************************************/
unsigned
GetDateArg (char *Str) {
	char Temp[20], *p, *t;
	int	Args[3], i ,j;

	strcpy(Temp, Str);
	p = Temp;
	for (i=0; i < 3; ++i) Args[i] = 0;
	j = strcspn(p, " -/");
	if (j >= strlen(p)) {
		for (j=2; j >= 0; --j) {
			Args[j] = atoi(&p[j * 2]);
			p[j*2] = '\0';
			}
		}
	else {
		for (j=0; j < 3; ++j) {
			t = &p[strcspn(p, " -/")];
			for (i=0; i < t - p; ++i) if (!isdigit(p[i])) break;
			if (i < t - p) Error("Invalid Date parameter: %s - %s", p, Str);
			Args[j] = atoi(p);
			if (*t != '\0') p = ++t;
			else break;
			}
		}
	if (Args[2] > 1980) Args[2] -= 1980;
	else Args[2] -= 80;
	return( (Args[2] << 9) | ((Args[0] & 0x0F) << 5) | (Args[1] & 0x1F) );
	}


