#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>
#ifdef __TURBOC__
	#include <dir.h>
#else
	#ifdef __POWERC__
		#include <dir.h>
	#else
		#ifdef M_I86
			#include <direct.h>
		#endif
	#endif
#endif

/***************************************************************************
 * RootPath  --  convert a pathname argument to root based cannonical form
 *
 * RootPath determines the current directory, appends the path argument (which
 * may affect which disk the current directory is relative to), and qualifies
 * "." and ".." references.  The result is a complete, simple, path name with
 * drive specifier.
 *
 * If the relative path the user specifies does not include a drive spec., the
 * default drive will be used as the base.  (The default drive will never be
 * changed.)
 *
 *  entry: RelativePath  -- pointer to the pathname to be expanded
 *         FullPath -- must point to a working buffer, see warning
 *   exit: FullPath -- the full path which results
 * return: true if an error occurs, false otherwise
 *
 * calls: getcurdir getdisk
 *
 * warning: FullPath must point to a working buffer large enough to hold the
 *          longest possible relative path argument plus the longest possible
 *          current directory path.
 *
 ***************************************************************************/
int
RootPath(char *RelativePath, char *FullPath) {
	extern char CurDir[];

    register char *Ptr;

/* Extract drive spec */

	if ((*RelativePath != '\0') && (RelativePath[1] == ':')) {
		strncpy(FullPath, RelativePath, 2);
        RelativePath += 2;
		}
    else {
		strncpy(FullPath, CurDir, 2);
		}
	FullPath[2] = '\0';

/* Append RelativePath to FullPath/base */

    if (*RelativePath == '\\' || *RelativePath == '/') {  /* RelativePath starts at base */
        strcpy(FullPath+2, RelativePath);
		}
	else if (*RelativePath != '\0') { 		/* must get base path first */
        FullPath[2] = '\\';
		if (FullPath[0] != CurDir[0]) {
#ifdef __TURBOC__
	        if (getcurdir( (int) (toupper(FullPath[0]) - '@'), &FullPath[3])) {
				return(1);           /* terrible error */
				}
#else
	#ifdef __POWERC__
	        if (getcurdir( (int) (toupper(FullPath[0]) - '@'), &FullPath[3])) {
				return(1);           /* terrible error */
				}
	#else
		#ifdef M_I86
			unsigned NumberOfDrives;

			_dos_setdrive( (int) (toupper(FullPath[0]) - '@'), &NumberOfDrives);
			getcwd(FullPath, sizeof(FullPath) - 1);
			_dos_setdrive( (int) (toupper(CurDir[0]) - '@'), &NumberOfDrives);
		#endif
	#endif
#endif
			}
		else strcpy(FullPath, CurDir);
        if ((*RelativePath != '\0') && (strlen(FullPath) > 3))
            strcat(FullPath, "\\");
        strcat(FullPath, RelativePath);
		}

/* Convert path to cannonical form */

	while ( (Ptr = strchr(FullPath, '/')) != NULL) *Ptr++ = '\\';
	Ptr = strrchr(FullPath, '\\');
	if (!strcmp(&Ptr[1], ".")) *Ptr = '\0';
	else if (!strcmp(&Ptr[1], "..")) {
		*Ptr = '\0';
		Ptr = strrchr(FullPath, '\\');
		*Ptr = '\0';
		}

    if ( (strlen(FullPath) == 3) && (FullPath[2] == '\\') ) FullPath[2] = '\0';

/* Shift to lower case */

    strlwr(FullPath);

    return(0);
	}
