/*
UFind - derived from:
	Find - Copyright (C) 1988, Russell Nelson
	by Don A. Williams

Copying for all purposes permitted so long as this copyright notice is
retained.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <conio.h>
#include <stdarg.h>
#include <io.h>
#include <dos.h>
#ifdef __TURBOC__
	#include <mem.h>
	#include <dir.h>
#else
	#ifdef __POWERC__
		#include <mem.h>
		#include <dir.h>
	#else
		#ifdef M_I86
			#include <memory.h>
			#include <direct.h>
		#endif
	#endif
#endif

#define MAIN 1

#include "queue.h"
#include "ufind.h"
#include "copymove.h"
#include "execute.h"

char *PrintNumber (long Number);

void
main (int argc, char *argv[]) {
	int l;

	fprintf(stderr, "UFind - Version %s:  %s\n", VERSION, VERSION_DATE);

	InitQueue(&PNames);
	Today = SysDate();
	SiDate = BeDate = Today;

	getcwd(CurDir, sizeof(CurDir));
	strlwr(CurDir);
	if (CurDir[strlen(CurDir)-1] == '\\') CurDir[strlen(CurDir)-1] = '\0';

	SwitChar = '-';

	if (argc > 1) GetArgs(&argv);

/* sanity checks */

	if ((AttrOn & AttrOff) != 0) Error("overlapping attr and ^attr");
	if ((IsAttrOn & IsAttrOff) != 0) Error("overlapping attr+ and attr-");
	if (PNames.Count == 0)  Enque(&PNames, CurDir);
	if (PrintFlag && QuietFlag)
		Error("Canot print and be silent at the same time");
	if (!CpFlag && !MvFlag && !RmFlag && !RmDirFlag	&& !QuietFlag
		&& !AttrOn && !AttrOff) PrintFlag = TRUE;

/* read in files */

	for (Ptr = PNames.Head; Ptr != NULL; Ptr = Ptr->Next) {
		BaseLength = strlen(Ptr->Body);
		FindFiles(Ptr->Body);
		if (RmDirFlag) {	/* special kludge to remove ourselves */
			l = strlen(Ptr->Body) - 1;
			if (Ptr->Body[l-1] != ':') {
				if ( (Ptr->Body[l] == '\\') || (Ptr->Body[l] == '/') )
					Ptr->Body[l] = '\0';
				if (YesNo("RMDIR %s", Ptr->Body))
					if (rmdir(Ptr->Body)) perror("    FAILED");
				}
			}
		}
	if (SumFlag) {
		printf("%s", PrintNumber(SumSize));
		printf(" bytes total in the files found.", SumSize);
		}
	}

/**************************************************************************
 * Error - Display an error message and exit with ERRORLEVEL = 1
 *************************************************************************/
void
Error (char *Format, ...) {
	va_list ArgPtr;

	printf("UFind: ");
	va_start(ArgPtr, Format);
	vprintf(Format, ArgPtr);
	va_end(ArgPtr);
	printf("\n");
	exit(1);
	}

/**************************************************************************
 * ParseAttributes -
 *************************************************************************/
int
ParseAttributes(char *Attributes) {
	int Result = 0;

	while (*Attributes != '\0') switch(tolower(*Attributes++)) {
		case 'a': Result |= ARCH;   break;
		case 'd': Result |= DIREC_;  break;
		case 'v': Result |= LABEL_;  break;
		case 's': Result |= SYSTEM_; break;
		case 'h': Result |= HIDDEN_; break;
		case 'r': Result |= RDONLY_; break;
		default: Error("unknown attribute letter, '%c'", Attributes[-1]);
		}
	return (Result);
	}

/**************************************************************************
 * FindFiles - Recursive function to walk the directory hierarchy, starting
 *	 		   with the specified directory, and call PerformAction() for
 *			   every file found.
 *************************************************************************/
void
FindFiles(char *BaseName) {
	char ThisName[64];
	struct DIR_BLK OurFF;

	sprintf(ThisName, "%s\\*.*", BaseName);
	if (!FIND_FIRST(ThisName, &OurFF, 0xFF)) do {

/* ignore . and .. */

		if (IsDir(OurFF.DIR_ATTR) && OurFF.DIR_FNAME[0] == '.') continue;

		strlwr(OurFF.DIR_FNAME);

		if (!NrFlag && IsDir(OurFF.DIR_ATTR)) {
			char newbase[64];
			sprintf(newbase, "%s\\%s", BaseName, OurFF.DIR_FNAME);
			FindFiles(newbase);
			}
		if ( Matches(BaseName, &OurFF) ) PerformAction(BaseName, &OurFF);
		}
	while (!FIND_NEXT(&OurFF));
	}

/**************************************************************************
 * PerformAction - Performs the opeerations requested on the command line
 *				   on the specified file (MS-DOS Directory Structure) in the
 *				   specified directory.
 *************************************************************************/
void
PerformAction(char *BaseName, struct DIR_BLK *FF) {

	static char *Months[] = {
		"Jan", "Feb", "Mar", "Apr", "May", "Jun",
		"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
		};
	char ThisN[64], *Destin;

	int ThisRmFlag = RmFlag;	/* in case we fail to move or copy it */

/* Build the complete pathname */

	strcpy(ThisN, BaseName);
	strcat(ThisN, "\\");
	strcat(ThisN, FF->DIR_FNAME);

/* Print action */

	if (!LsFlag && !LdFlag && PrintFlag && !PromptFlag && !NFlag) puts(ThisN);
	else if ( (LsFlag || LdFlag) && PrintFlag) {
		FDate.Date = FF->DIR_DATE;
		FileTime.time = FF->DIR_TIME;
		printf("%c%c%c%c%c%c",
			FF->DIR_ATTR & ARCH   ? 'a' : '-',	/* for the archive bit being on */
			FF->DIR_ATTR & DIREC_  ? 'd' : '-',	/* for the directory bit being on */
			FF->DIR_ATTR & LABEL_  ? 'v' : '-',	/* for the file being a volume label */
			FF->DIR_ATTR & SYSTEM_ ? 's' : '-',	/* the for file being a system file */
			FF->DIR_ATTR & HIDDEN_ ? 'h' : '-',	/* for the file being hidden */
			FF->DIR_ATTR & RDONLY_ ? '-' : 'w'	/* for the file being read-only */
			);
		printf(" %16s ", PrintNumber(FF->DIR_FSIZE));
		if (LsFlag) printf("%2u-%3s-%02u ",
			FDate.d.day,
			Months[FDate.d.month - 1],
			FDate.d.year+80);
		else printf("%02u-%02u-%02u ",
			FDate.d.year+80,
			FDate.d.month,
			FDate.d.day);
		printf("%2u:%02u %s\n",
			FileTime.t.hour,
			FileTime.t.minute,
			ThisN);
		}

/* attribute action */

	if (ExecFlag) ExecuteCommand(ThisN);
				  
	if ((AttrOn != 0 || AttrOff != 0) && YesNo("ATTR+ %s", ThisN)) {
		unsigned char NewAttr = (FF->DIR_ATTR | AttrOn) & ~ AttrOff;
		if (chmod(ThisN, NewAttr) < 0) Error("can't set attributes on file '%s' to %u", ThisN, NewAttr);
		FF->DIR_ATTR = NewAttr;
		}

	if ((CpFlag || MvFlag || BackupFlag)
		 && (strcmp(ThisN,
			 		(Destin = strcat(OtherName(BaseName , ToDir), FF->DIR_FNAME))))
		 && !(FF->DIR_ATTR & DIREC_) ) {
		char *Result, Src[64], Dest[64];

		RootPath(ThisN, Src);
		RootPath(Destin, Dest);
		if (strcmp(ThisN, Destin))
            if (*(Result = CopyMove(ThisN, Destin, MvFlag, 0)) == '?')
				Error(Result);
		if (MvFlag) ThisRmFlag = FALSE;
		}

	if ( ThisRmFlag && !IsDir(FF->DIR_ATTR)
		&& YesNo("RM %s", ThisN) ) if ( unlink(ThisN) )
			perror("    FAILED");

	if (RmDirFlag && IsDir(FF->DIR_ATTR)) {
		if (YesNo("RMDIR %s", ThisN)) if (rmdir(ThisN))
			perror("    FAILED");
		}

	if (SumFlag) SumSize += FF->DIR_FSIZE;
	}

/**************************************************************************
 * Matches -- return TRUE if this file matches all the criteria.
 *************************************************************************/
int
Matches (char *BaseName, struct DIR_BLK *FF) {
	char ThisN[64];
	int i;
	int AMatch(char *Str, char *Pat);

	strcpy(ThisN, BaseName);
	strcat(ThisN, FF->DIR_FNAME);

	for (i = 0; i < IncNamesCnt; i++) {
		if (AMatch(FF->DIR_FNAME, IncNames[i])) break;
		}
	if (IncNamesCnt > 0 && i == IncNamesCnt) return (FALSE);

	for (i = 0; i < ExcNamesCnt; i++) {
		if (AMatch(FF->DIR_FNAME, ExcNames[i])) return (FALSE);
		}

	for (i = 0; i < CntIncFNames; i++) {
		if (AMatch(ThisN, IncFNames[i])) break;
		}
	if (CntIncFNames > 0 && i == CntIncFNames) return (FALSE);

	for (i = 0; i < CntExcFNames; i++) {
		if (AMatch(ThisN, ExcFNames[i])) return (FALSE);
		}

	if (IsAttrOn != 0 && (FF->DIR_ATTR & IsAttrOn) == 0) return (FALSE);
	if (IsAttrOff != 0 && (FF->DIR_ATTR & IsAttrOff) != 0) return (FALSE);

	if (SizeFlag == '+' && FF->DIR_FSIZE < SizeSize) return (FALSE);
	if (SizeFlag == '-' && FF->DIR_FSIZE > SizeSize) return (FALSE);
	if (SizeFlag == '=' && FF->DIR_FSIZE != SizeSize) return (FALSE);

	if (ExistsFlag != 0) {
		char ThatFn[64];
		struct DIR_BLK ThatFF;
		int FoundIt;

		sprintf(ThatFn, "%s\\%s", ExistsName, FF->DIR_FNAME);
		FoundIt = !FIND_FIRST(ThatFn, &ThatFF, 0);

		if (ExistsFlag > 0 && FoundIt == 0) return (FALSE);	/* doesn't exist and we want it to. */
		if (ExistsFlag < 0 && FoundIt != 0) return (FALSE);	/* does exist and we don't want it to. */
		}
	if (DtStyle != none && DtDir) {
		char ThatFn[64];
		struct DIR_BLK ThatFF;
		int FoundIt;

		sprintf(ThatFn, "%s\\%s", DtName, FF->DIR_FNAME);
		FoundIt = FIND_FIRST(ThatFn, &ThatFF, 0);	/* doesn't exist */

		if (DtStyle != none && FoundIt != 0) return (FALSE); /* doesn't exist -- we can't determine older or newer */

		DtDate = ThatFF.DIR_DATE;
		DtTime = ThatFF.DIR_TIME;
		}
	switch(DtStyle) {
		case none:
			break;
		case LE:
			if (!(CompareDates(FF->DIR_DATE, DtDate, FF->DIR_TIME, DtTime)
				<= 0)) return (FALSE);
			break;
		case LT:
			if (!(CompareDates(FF->DIR_DATE, DtDate, FF->DIR_TIME, DtTime)
				< 0)) return (FALSE);
			break;
		case GE:
			if (!(CompareDates(FF->DIR_DATE, DtDate, FF->DIR_TIME, DtTime)
				>= 0)) return (FALSE);
			break;
		case GT:
			if (!(CompareDates(FF->DIR_DATE, DtDate, FF->DIR_TIME, DtTime)
				> 0)) return (FALSE);
			break;
		case LEDATE:
			if (FF->DIR_DATE > DtDate) return (FALSE);
			break;
		case LTDATE:
			if (FF->DIR_DATE >= DtDate) return (FALSE);
			break;
		case GEDATE:
			if (FF->DIR_DATE < DtDate) return (FALSE);
			break;
		case GTDATE:
			if (FF->DIR_DATE <= DtDate) return (FALSE);
			break;
		case SI:
			if (FF->DIR_DATE < SiDate) return (FALSE);
			break;
		case BE:
			if (FF->DIR_DATE >= BeDate) return (FALSE);
			break;
		case SIBE:
			if ( (FF->DIR_DATE >= BeDate)
				|| (FF->DIR_DATE < SiDate) ) return(FALSE);
			break;
		}
	if (CompFlag) {
		struct DIR_BLK ThatFF;
		char *Destin;
		int Identical;

		Destin = OtherName(BaseName, CompDir);
		strcat(Destin, FF->DIR_FNAME);
		if (NewSwt) Identical = !(FIND_FIRST(Destin, &ThatFF, 0));
		else Identical = !FIND_FIRST(Destin, &ThatFF, 0)
				&& FF->DIR_FSIZE == ThatFF.DIR_FSIZE
				&& FF->DIR_DATE == ThatFF.DIR_DATE
				&& FF->DIR_TIME == ThatFF.DIR_TIME;
		if (CompFlag > 0 && Identical == 0) return (FALSE);	/* don't compare and we want them to. */
		if (CompFlag < 0 && Identical != 0) return (FALSE);	/* do compare and we don't want them to. */
		}

	return (TRUE);
	}

/**************************************************************************
 * YesNo - Displays a specified prompt message and, if the prompt flag is
 *		   ON, request a YES/NO response.  Returns TRUE if the prompt flag
 *		   is OFF or if the response is YES.
 *************************************************************************/
int
YesNo (char *Format, ...) {
	va_list ArgPtr;
	char ch;

	if (QuietFlag) return(TRUE);

	va_start(ArgPtr, Format);
	vfprintf(stderr, Format, ArgPtr);
	va_end(ArgPtr);

/* if no prompting, always do it */

	if (!PromptFlag && !NFlag) {
		fprintf(stderr, "\n");
		fflush(stderr);
		return(TRUE);
		}

	fprintf(stderr, " (Y/N)? ");
	for (;;) {
		ch = (NFlag) ? 'N': (char) getch();
		switch(toupper(ch)) {
			case 'N':
				fprintf(stderr, "No \n");
				fflush(stderr);
				return (FALSE);

			case 'Y':
				fprintf(stderr, "Yes\n");
				fflush(stderr);
				return (TRUE);

			default:
				fprintf(stderr, "\a");
				fflush(stderr);
				break;
			}
		}
	}

char *
OtherName(char *BaseName, char *ToDir) {
	extern int BaseLength;

	static char Destin[64];

	strcpy(Destin, ToDir);
	if (TFlag) {
		strcat(Destin, &BaseName[BaseLength]);
		}
	EnsureExists(Destin);
	return( strcat(Destin, "\\") );
	}

/**************************************************************************
 * CompareDates -- return -1 if a is older than b, 0 if a = b, 1 if a is
 *				   newer than b
 *************************************************************************/
int
CompareDates(int ADate, int BDate, int ATime, int BTime) {

	if (ADate > BDate) return(1);
	else if (ADate < BDate) return(-1);
	else if (ATime > BTime) return(1);
	else if (ATime < BTime) return(-1);
	else return(0);
	}

/****************************************************************************
* PrintNumber -- Print a long number, inserting ","s where needed.
****************************************************************************/
char *
PrintNumber (long Number) {
	static char Temp[20];
	int l, i;

	sprintf(Temp, "%lu", Number);
	l = strlen(Temp);
	for (i=0; i < l; ++i) {
		if ( i && ((i) % 3) == 0) {
			memmove(&Temp[l-i+1], &Temp[l-i], strlen(&Temp[l-i])+1);
			Temp[l-i] = ',';
			}
		}
	return(Temp);
	}

/****************************************************************************
 *  Revision History
 ***************************************************************************/
/* Version 1.10 - February 25, 1988                                        */
/* Fixed the "-p" option if "-print" is NOT specified.                     */
/* Repaired the determination of the end of "-^name" and similar options.
	The bug was introduced in version 1.00 of UFIND.                       */

/* Version 1.10 - February 22, 1988                                        */
/* Added "-ld" option to change the date of the "-ls" display from "military"
	format to "international" format to allow sorting on date.             */
/* Repaired the path determination for the "-t" option                     */
/* Added error messages when the "-rm" or "-rmdir" options fail            */

/* 02-14-88          Moved Revison History to end of source.  DAW */
/* 02-14-88			 Added "QuietFlag" and changed operation of PrintFlag to
                     suppress "extra" printing.               DAW */
/* 02-14-88			 Rearranged code in UFIND.C to include modules in order
                     of use.                                  DAW */
/* 02-14-88          Created UFIND.H containing function prototypes for all
					 functions in UFIND.C                     DAW */
/* 02-14-88			 Added "Version" identification           DAW */
/* 02-14-88          Rewrote "exec" option					  DAW */
/* 02-14-88			 Added "-si" and "-be" options			  DAW */
/* 02-14-88			 Fixed "-mv" and "-cp" options to NOT attempt the
					 operation if the source and destination directories are
					 the same.								  DAW */
/* 02-14-88			 Enabled the "prompt" for "rm" amd "rmdir" as well as
					 "mv" and "cp".							  DAW */
/* 02-14-88			 Fixed "rmdir" to remove the base level directory if it
					 is empty.								  DAW */

/* History:770,1 */
/* 01-30-88 15:34:13 if the current directory is '/', don't use './' */
/* 10-26-87 22:45:43 add EnsureExists() */
/* 10-25-87 23:54:40 always say what we're doing if they ask to print. */
/* 10-25-87 23:41:15 add backup code. */
/* 10-25-87 23:32:53 call RootPath() before Cp(). */
/* 10-25-87 23:28:44 if we're removing directories, be sure to remove the top level (if tempty) */
/* 10-25-87 23:12:37 add Cp() */
/* 10-13-87 22:35:37 Allow -<multiple options> */
/* 10-12-87 10:35:56 rearrange PerformAction so that it makes more sense. */
/* 10-11-87 23:06:52  */
/* 10-11-87 23:05:36 add the code for -comp */
/* 10-09-87 08:24:42  */
/* 10-07-87 23:23:17 implement rmdir */
/* 10-07-87 22:07:17 add -size. */
/* 10-07-87 00:33:53 Add fnames */
/* 10-07-87 00:09:26 move AMatch out. */
/* 10-07-87 00:05:56 switch AMatch() around. */
/* 10-06-87 23:53:41 change the switch names. */
/* 10-06-87 23:47:00 add AMatch() */
/* 10-04-87 22:52:42 add exists_flag. */
/* 10-02-87 23:49:58 rearrange the yesno switch. */
/* 10-02-87 23:39:10 fflush after fprintf */
/* 10-01-87 23:47:55 add arguments to Error(). */
/* 09-30-87 03:22:58 use a table of strings for the switch names. */
/* 09-30-87 00:58:14 remove sets, remove interactive operation. */
/* 08-11-87 22:22:57 add Fopendir flag to File */
/* 07-29-87 22:57:28 add LT, etc... */
/* 07-29-87 12:59:15 write and use compare_dates */
/* 07-28-87 12:45:11 put nr_flag in the correct place. */
/* 07-28-87 08:04:14 add multiple pathnames readin. */
/* 07-28-87 07:43:36 add -newer and -older */
/* 07-27-87 00:31:50 work on Cp() a little more. */
/* 07-21-87 22:59:28 start adding Cp() */
/* 07-21-87 00:08:28 add ctrlbreak checking. */
/* 07-21-87 00:02:07 add sets */
/* 07-19-87 23:32:07 add multiple filename support */
/* 07-19-87 23:16:09 make findFile use possibly relative pathnames. */
/* 07-19-87 22:30:39 add -attr. */
/* 07-19-87 22:25:23 in parsing -name, don't ignore the following argument. */
/* 07-19-87 22:17:10 change 'conflicting' to 'overlapping' in attr-, attr+ error message. */
/* 07-19-87 22:11:09 set the attributes in the in-memory copy also. */
/* 07-19-87 00:34:06 use the attributes that we have in Fattr, don't get them via _chmod. */
/* 07-19-87 00:30:30 put usage before usage(). */
/* 07-19-87 00:29:39 test to see if _chmod is succeeding. */
/* 07-18-87 23:52:48 get rid of ambiguous operator warnings. */
/* 07-18-87 23:50:51 add attr+ and attr- */
/* 07-18-87 23:13:46 use month-1 when accessing Months[]. */
/* 07-18-87 23:10:45 make time bitfields into unsigned's, and add 80 to raw year. */
/* 07-18-87 23:07:16 fix more union problems. */
/* 07-18-87 23:01:45 fix syntax errors in the unions. */
/* 07-18-87 22:56:20 finish adding -ls */
/* 07-18-87 14:44:50 add -ls */
/* 07-18-87 14:18:29 include <process.h> for exit() */
/* 07-18-87 12:35:00 add a linefeed after the cgets. */
/* 07-18-87 12:33:02 fix the checking for attribute bits. */
/* 07-18-87 12:31:54 add a prompt string. */
/* 07-18-87 12:30:43 include <ctype.h> */
/* 07-18-87 12:30:12 in writelist, write the entire filename, not just the last pointer. */
/* 07-18-87 12:25:18 include <conio.h> */
/* 07-18-87 12:20:53 finish adding command line interface. */
/* 07-18-87 11:43:08 start to add a command line interface. */
/* 07-17-87 23:43:49 add a usage function. */
/* 07-17-87 23:31:01  */
/* 07-17-87 13:23:36 match for filenames correctly. */
/* 07-17-87 13:21:17 add the Matches() function. */
/* 07-17-87 00:05:30 merge ext and name back into name. */
/* 07-16-87 23:26:31 include <string.h> */
/* 07-16-87 23:19:59 include <alloc.h> */
/* 07-16-87 22:54:57 store them in a global list. */
/* 07-16-87 22:39:36 print out what we're doing as we go. */
/* 07-16-87 22:31:48 don't recurse down '.' or '..'. */
/* 07-16-87 21:48:04 first creation */

